// rdparse2_test.cpp
// VERSION 2
// Glenn G. Chappell
// 21 Feb 2013
//
// For CS 331 Spring 2013
// Test program for class RDParse2
// Also required lex2.h/.cpp
// Used in Assignment 3, Exercise A

// Includes for code to be tested
#include "rdparse2.h"   // For class RDParse2
#include "rdparse2.h"   // Double inclusion test

// Includes for testing package & code common to all test programs
#include <iostream>     // for std::cout, std::endl, std::cin
#include <string>       // for std::string
#include <stdexcept>    // for std::runtime_error

// Additional includes for this test program
#include <stdexcept>    // for std::exception
#include <sstream>      // for std::ostringstream, std::istringstream

// If the constant SINGLE_TEST is defined, then this program terminates
// after the first failed test (if any). Comment out the following line
// to have the program run all tests, regardless.
#define SINGLE_TEST


// ************************************************************************
// Testing Package:
//     Class Tester - For Tracking Tests
// ************************************************************************


// class Tester
// For extremely simple unit testing.
// Keeps track of number of tests and number of passes.
// Use test (with success/failure parameter) to do a test.
// Get results with numTests, numPassed, numFailed, allPassed.
// Restart testing with reset.
// Invariants:
//     countTests_ == number of tests (calls to test) since last reset.
//     countPasses_ == number of times function test called with true param
//      since last reset.
//     0 <= countPasses_ <= countTests_.
//     tolerance_ >= 0.
class Tester {

// ***** Tester: ctors, dctor, op= *****
public:

    // Default ctor
    // Sets countTests_, countPasses_ to zero, tolerance_ to given value
    // Pre: None.
    // Post:
    //     numTests == 0, countPasses == 0, tolerance_ == abs(theTolerance)
    // Does not throw (No-Throw Guarantee)
    Tester(double theTolerance = 0.0000001)
        :countTests_(0),
         countPasses_(0),
         tolerance_(theTolerance >= 0 ? theTolerance : -theTolerance)
    {}

    // Compiler-generated copy ctor, copy op=, dctor are used

// ***** Tester: general public functions *****
public:

    // test
    // Handles single test, param indicates pass/fail
    // Pre: None.
    // Post:
    //     countTests_ incremented
    //     countPasses_ incremented if (success)
    //     Message indicating test name (if given)
    //      and pass/fail printed to cout
    // Does not throw (No-Throw Guarantee)
    //  - Assuming exceptions have not been turned on for cout.
    void test(bool success,
              const std::string & testName = "")
    {
        ++countTests_;
        if (success) ++countPasses_;

        std::cout << "    ";
        if (testName != "")
        {
            std::cout << "Test: "
                      << testName
                      << " - ";
        }
        std::cout << (success ? "passed" : "********** FAILED **********")
                  << std::endl;
    }

    // ftest
    // Does single floating-point test.
    // Tests passes iff difference of first two values is <= tolerance.
    // Pre: None.
    // Post:
    //     countTests_ incremented
    //     countPasses_ incremented if (abs(val1-val2) <= tolerance_)
    //     Message indicating test name (if given)
    //      and pass/fail printed to cout
    // Does not throw (No-Throw Guarantee)
    void ftest(double val1,
               double val2,
               const std::string & testName = "")
    { test(val1-val2 <= tolerance_ && val2-val1 <= tolerance_, testName); }

    // reset
    // Resets *this to default constructed state
    // Pre: None.
    // Post:
    //     countTests_ == 0, countPasses_ == 0
    // Does not throw (No-Throw Guarantee)
    void reset()
    {
        countTests_ = 0;
        countPasses_ = 0;
    }

    // numTests
    // Returns the number of tests that have been done since last reset 
    // Pre: None.
    // Post:
    //     return == countTests_
    // Does not throw (No-Throw Guarantee)
    int numTests() const
    { return countTests_; }

    // numPassed
    // Returns the number of tests that have passed since last reset
    // Pre: None.
    // Post:
    //     return == countPasses_
    // Does not throw (No-Throw Guarantee)
    int numPassed() const
    { return countPasses_; }

    // numFailed
    // Returns the number of tests that have not passed since last reset
    // Pre: None.
    // Post:
    //     return + countPasses_ == numTests_
    // Does not throw (No-Throw Guarantee)
    int numFailed() const
    { return countTests_ - countPasses_; }

    // allPassed
    // Returns true if all tests since last reset have passed
    // Pre: None.
    // Post:
    //     return == (countPasses_ == countTests_)
    // Does not throw (No-Throw Guarantee)
    bool allPassed() const
    { return countPasses_ == countTests_; }

    // setTolerance
    // Sets tolerance_ to given value
    // Pre: None.
    // Post:
    //     tolerance_ = abs(theTolerance)
    // Does not throw (No-Throw Guarantee)
    void setTolerance(double theTolerance)
    { tolerance_ = (theTolerance >= 0 ? theTolerance : -theTolerance); }

// ***** Tester: data members *****
private:

    int countTests_;    // Number of tests done since last reset
    int countPasses_;   // Number of tests passed since last reset
    double tolerance_;  // Tolerance for floating-point near-equality tests

};  // end class Tester


// ************************************************************************
// Testing Package:
//     Class TypeCheck - Helper Class for Type Checking
// ************************************************************************


// class TypeCheck
// This class exists in order to have static member function check, which
// takes a parameter of a given type, by reference. Objects of type
// TypeCheck<T> cannot be created.
// Usage:
//     TypeCheck<MyType>::check(x)
//     returns true if the type of x is (MyType) or (const MyType),
//     otherwise false.
// Invariants: None.
// Requirements on Types: None.
template<typename T>
class TypeCheck {

private:

    // Uncopyable class. Do not define copy ctor, copy assn.
    TypeCheck(const TypeCheck &);
    TypeCheck<T> & operator=(const TypeCheck &);

    // Compiler-generated dctor is used (but irrelevant).

public:

    // check
    // The function and function template below simulate a single function
    // that takes a single parameter, and returns true iff the parameter has
    // type T or (const T).

    // check (reference-to-const T)
    // Pre: None.
    // Post:
    //     Return is true.
    // Does not throw (No-Throw Guarantee)
    static bool check(const T & param)
    { return true; }

    // check (reference-to-const non-T)
    // Pre: None.
    // Post:
    //     Return is false.
    // Requirements on types: None.
    // Does not throw (No-Throw Guarantee)
    template <typename OtherType>
    static bool check(const OtherType & param)
    { return false; }

};  // End class TypeCheck


// ************************************************************************
// Testing Package:
//     Class Counter - Helper Class for Counting Calls & Objects, Throwing
// ************************************************************************


// class Counter
// Item type for counting ctor, dctor, op= calls, counting existing
//  objects, and possibly throwing on copy. Has operator< (which always
//  returns false), allowing it to be the value type of a sorted container.
// If static member copyThrow_ is set, then copy ctor and copy assn throw
//  std::runtime_error. Exception object's "what" member is set to "C" by
//  the copy ctor and "A" by copy assn.
// Increments static data member ctorCount_ on default construction and
//  successful copy construction. Increments static data member assnCount_
//  on successful copy assignment. Increments static data member
//  dctorCount_ on destruction.
// Increments static data member existing_ on construction, and decrements
//  it on destruction.
// Static data member maxExisting_ is highest value of existing_ since last
//  reset, or start of program if reset has never been called.
// Invariants:
//     Counter::existing_ is number of existing objects of this class.
//     Counter::ctorCount_ is number of successful ctor calls since
//      most recent call to reset, or start of program if reset has never
//      been called.
//     Counter::dctorCount_ is (similarly) number of dctor calls.
//     Counter::assnCount_ is (similarly) number of copy assn calls.
//     Counter::maxExisting_ is (similarly) highest value existing_ has
//      assumed.
class Counter {

// ***** Counter: Ctors, dctor, op= *****
public:

    // Default ctor
    // Pre: None.
    // Post:
    //     (ctorCount_ has been incremented.)
    //     (existing_ has been incremented.)
    // Does not throw (No-Throw Guarantee)
    Counter()
    {
        ++existing_;
        if (existing_ > maxExisting_)
            maxExisting_ = existing_;
        ++ctorCount_;
    }

    // Copy ctor
    // Throws std::runtime_error if copyThrow_.
    // Pre: None.
    // Post:
    //     (ctorCount_ has been incremented.)
    //     (existing_ has been incremented.)
    // May throw std::runtime_error
    // Strong Guarantee
    Counter(const Counter & other)
    {
        if (copyThrow_)
            throw std::runtime_error("C");
        ++existing_;
        if (existing_ > maxExisting_)
            maxExisting_ = existing_;
        ++ctorCount_;
    }

    // Copy assignment
    // Throws std::runtime_error if copyThrow_.
    // Pre: None.
    // Post:
    //     Return value is *this.
    //     (assnCount_ has been incremented.)
    // May throw std::runtime_error
    // Strong Guarantee
    Counter & operator=(const Counter & rhs)
    {
        if (copyThrow_)
            throw std::runtime_error("A");
        ++assnCount_;
        return *this;
    }

    // Dctor
    // Pre: None.
    // Post:
    //     (dctorCount_ has been incremented.)
    //     (existing_ has been decremented.)
    // Does not throw (No-Throw Guarantee)
    ~Counter()
    {
        --existing_;
        ++dctorCount_;
    }

// ***** Counter: Functions dealing with count *****
public:

    // reset
    // Pre: None.
    // Post:
    //     maxExisting_ == existing_.
    //     ctorCount_ == 0.
    //     dctorCount_ == 0.
    //     assnCount_ == 0.
    //     copyThrow_ == shouldThrow.
    // Does not throw (No-Throw Guarantee)
    static void reset(bool shouldThrow = false)
    {
        maxExisting_ = existing_;
        ctorCount_ = 0;
        dctorCount_ = 0;
        assnCount_ = 0;
        copyThrow_ = shouldThrow;
    }

    // getExisting
    // Pre: None.
    // Post:
    //     return == existing_.
    // Does not throw (No-Throw Guarantee)
    static int getExisting()
    { return existing_; }

    // getMaxExisting
    // Pre: None.
    // Post:
    //     return == maxExisting_.
    // Does not throw (No-Throw Guarantee)
    static int getMaxExisting()
    { return maxExisting_; }

    // getCtorCount
    // Pre: None.
    // Post:
    //     return == ctorCount_.
    // Does not throw (No-Throw Guarantee)
    static int getCtorCount()
    { return ctorCount_; }

    // getDctorCount
    // Pre: None.
    // Post:
    //     return == dctorCount_.
    // Does not throw (No-Throw Guarantee)
    static int getDctorCount()
    { return dctorCount_; }

    // getAssnCount
    // Pre: None.
    // Post:
    //     return == assnCount_.
    // Does not throw (No-Throw Guarantee)
    static int getAssnCount()
    { return assnCount_; }

    // setCopyThrow
    // Pre: None.
    // Post:
    //     copyThrow_ == shouldThrow
    // Does not throw (No-Throw Guarantee)
    static void setCopyThrow(bool shouldThrow)
    { copyThrow_ = shouldThrow; }

// ***** Counter: Data Members *****
private:

    static int existing_;     // # of existing objects
    static int maxExisting_;  // Max # of existing objects
    static int ctorCount_;    // # of successful (non-throwing) ctor calls
    static int dctorCount_;   // # of dctor calls
    static int assnCount_;    // # of successful (non-throwing) copy = calls
    static bool copyThrow_;   // true if copy operations (ctor, =) throw

};  // End class Counter

// Definition of static data member of class Counter
int Counter::existing_ = 0;
int Counter::maxExisting_ = 0;
int Counter::ctorCount_ = 0;
int Counter::dctorCount_ = 0;
int Counter::assnCount_ = 0;
bool Counter::copyThrow_ = false;


// operator< (Counter)
// Dummy-ish operator<, forming a strict weak order for Counter class
// Returns false (which is legal for a strict weak order; all objects of
//  type Counter are equivalent).
// Pre: None.
// Post:
//     Return value == false.
// Does not throw (No-Throw Guarantee)
bool operator<(const Counter & a,
               const Counter & b)
{ return false; }


// ************************************************************************
// Test Suite Functions
// ************************************************************************


// lex_out
// Print lexeme
void lex_out(const Lex2::Lexeme & it)
{
    std::cout << "        ";
    switch (it.second)
    {
        case Lex2::NONE:
            std::cout << "NONE";
            break;
        case Lex2::ILLEGAL:
            std::cout << "ILLEGAL";
            break;
        case Lex2::KEY:
            std::cout << "KEY";
            break;
        case Lex2::ID:
            std::cout << "ID";
            break;
        case Lex2::OP:
            std::cout << "OP";
            break;
        case Lex2::PUNCT:
            std::cout << "PUNCT";
            break;
        case Lex2::INT:
            std::cout << "INT";
            break;
        case Lex2::FLOAT:
            std::cout << "FLOAT";
            break;
        default:
            std::cout << "<undefined token>";
    }
    std::cout << " [" << it.first << "]" << std::endl;
}


// parse_it
// Test class RDParse2 output. Is given initialized object,
// expected results. Ignores done, correctOutput if parsed is false.
// Pre: None.
// Post:
//     Pass/fail status of tests have been registered with t.
//     Appropriate messages have been printed to cout.
// If SINGLE_TEST not defined:
//     Does not throw (No-Throw Guarantee)
// If SINGLE_TEST defined:
//     Throws std::exception on test failure; Basic Guarantee
void parse_it(Tester & t,
              RDParse2 & parser,
              const std::string & testName,
              bool parsed,
              bool done = true,
              const std::string & correctOutput="")
{
    int errorp;  // 0: no error - should parse; did
                 // 1: no error - should not parse; did not
                 // 2: should parse; did not
                 // 3: should not parse; did
    int errord = 0;
                 // Only valid if errorp == 0 (else will be 0)
                 // 0: no error - should be done; was
                 // 1: no error - should not be done; was not
                 // 2: should be done; was not
                 // 3: should not be done; was
    int erroro = 0;
                 // Only valid if errorp == 0 (else will be 0)
                 // 0: no error
                 // 1: incorrect output

    std::string actualOutput;
    bool actualParsed = parser.parse(actualOutput);
    errorp = (parsed ? 0 : 3) ^ (actualParsed ? 0 : 2);
        // See above for values
    bool actualDone = parser.done();
    if (errorp == 0)  // parsed && actualParsed
    {
        errord = (done ? 0 : 3) ^ (actualDone ? 0 : 2);
            // See above for values
        erroro = (correctOutput != actualOutput);
    }

    bool good = (errorp <= 1 && errord <= 1 && erroro == 0);
    t.test(good, testName);
    if (good)
        return;

    if (errorp == 2)
    {
        std::cout << "      Correct program was flagged as incorrect"
                  << std::endl;
    }
    else if (errorp == 3)
    {
        std::cout << "      Incorrect program was not flagged as incorrect"
                  << std::endl;
    }
    else
    {
        if (errord == 2)
        {
            std::cout << "      Extra chars after end erroneously flagged"
                      << std::endl;
        }
        else if (errord == 3)
        {
            std::cout << "      Extra chars after end not flagged"
                      << std::endl;
        }

        if (erroro)
        {
            std::cout << "      Output of program was incorrect"
                      << std::endl;
            std::cout << "      Expected output:" << std::endl;
            std::cout << correctOutput;
            std::cout << std::endl;
            std::cout << "      Returned output:" << std::endl;
            std::cout << actualOutput;
            std::cout << std::endl;
        }
    }

#ifdef SINGLE_TEST
    throw std::exception();
#endif //#ifdef SINGLE_TEST
}


// parse_it_precision
// Test class RDParse2 output. Is given initialized object,
// expected results. Ignores done, correctVal if parsed is false.
// Pre: None.
// Post:
//     Pass/fail status of tests have been registered with t.
//     Appropriate messages have been printed to cout.
// If SINGLE_TEST not defined:
//     Does not throw (No-Throw Guarantee)
// If SINGLE_TEST defined:
//     Throws std::exception on test failure; Basic Guarantee
void parse_it_precision(Tester & t,
                        RDParse2 & parser,
                        const std::string & testName,
                        bool parsed,
                        bool done = true,
                        double correctVal = 0.)
{
    int errorp;  // 0: no error - should parse; did
                 // 1: no error - should not parse; did not
                 // 2: should parse; did not
                 // 3: should not parse; did
    int errord = 0;
                 // Only valid if errorp == 0 (else will be 0)
                 // 0: no error - should be done; was
                 // 1: no error - should not be done; was not
                 // 2: should be done; was not
                 // 3: should not be done; was

    std::string actualOutput;
    bool actualParsed = parser.parse(actualOutput);
    errorp = (parsed ? 0 : 3) ^ (actualParsed ? 0 : 2);
        // See above for values
    bool actualDone = parser.done();
    double actualVal;
    if (errorp == 0)  // parsed && actualParsed
    {
        errord = (done ? 0 : 3) ^ (actualDone ? 0 : 2);
            // See above for values
        std::istringstream iss(actualOutput);
        iss >> actualVal;
    }

    bool good = (errorp <= 1 && errord <= 1);
    if (good)
    {
        t.setTolerance(0.00001);
        t.ftest(correctVal, actualVal, testName + " (precision test)");
        return;
    }

    t.test(false, testName);

    if (errorp == 2)
    {
        std::cout << "      Correct program was flagged as incorrect"
                  << std::endl;
    }
    else if (errorp == 3)
    {
        std::cout << "      Incorrect program was not flagged as incorrect"
                  << std::endl;
    }
    else
    {
        if (errord == 2)
        {
            std::cout << "      Extra chars after end erroneously flagged"
                      << std::endl;
        }
        else if (errord == 3)
        {
            std::cout << "      Extra chars after end not flagged"
                      << std::endl;
        }
    }

#ifdef SINGLE_TEST
    throw std::exception();
#endif //#ifdef SINGLE_TEST
}


// test_class_RDParse2_simple
// Test suite for class RDParse2, simple inputs
// Pre: None.
// Post:
//     Pass/fail status of tests have been registered with t.
//     Appropriate messages have been printed to cout.
// If SINGLE_TEST not defined:
//     Does not throw (No-Throw Guarantee)
// If SINGLE_TEST defined:
//     Throws std::exception on test failure; Basic Guarantee
void test_class_RDParse2_simple(Tester & t)
{
    std::cout << "Test Suite: class RDParse2 - simple inputs"
              << std::endl;

    RDParse2 parser;
    bool parsed;
    bool done;
    std::string output;

    parser.set("");
    parsed = false;
    parse_it(t, parser, "empty string", parsed);

    parser.set("begin\nend");
    parsed = true;
    done = true;
    output = "";
    parse_it(t, parser, "empty correct program", parsed, done, output);

    parser.set("begin\nprint -3.2\nend");
    parsed = true;
    done = true;
    output = "-3.2\n";
    parse_it(t, parser, "single print", parsed, done, output);

    parser.set("begin\nabc = 2\nend");
    parsed = true;
    done = true;
    output = "";
    parse_it(t, parser, "single assign", parsed, done, output);

    parser.set("begin\nprint -3.2\nprint 5\nend");
    parsed = true;
    done = true;
    output = "-3.2\n5\n";
    parse_it(t, parser, "double print", parsed, done, output);

    parser.set("begin\nend x");
    parsed = true;
    done = false;
    output = "";
    parse_it(t, parser, "empty program, chars past end", parsed, done, output);

    parser.set("begin\nprint -3.2\nend x");
    parsed = true;
    done = false;
    output = "-3.2\n";
    parse_it(t, parser, "print, chars past end", parsed, done, output);

    parser.set("begin");
    parsed = false;
    parse_it(t, parser, "begin only", parsed);

    parser.set("end");
    parsed = false;
    parse_it(t, parser, "end only", parsed);

    parser.set("print 1.2");
    parsed = false;
    parse_it(t, parser, "print only", parsed);

    parser.set("begin\nprint\nend");
    parsed = false;
    parse_it(t, parser, "print w/o argument", parsed);

    parser.set("begin\nbegix=1.34\nprint begix\nend");
    parsed = true;
    done = true;
    output = "1.34\n";
    parse_it(t, parser, "assign & print", parsed, done, output);

    parser.set("begin\nbegin=1.34\nprint begin\nend");
    parsed = false;
    parse_it(t, parser, "reserved word as identifier", parsed);

    parser.set("/**/begin/**/\n/**/begix/*a*/=1.34\nprint/**/begix\nend/**/");
    parsed = true;
    done = true;
    output = "1.34\n";
    parse_it(t, parser, "Multiple comments", parsed, done, output);
}


// test_class_RDParse2_ops
// Test suite for class RDParse2, operators
// Pre: None.
// Post:
//     Pass/fail status of tests have been registered with t.
//     Appropriate messages have been printed to cout.
// If SINGLE_TEST not defined:
//     Does not throw (No-Throw Guarantee)
// If SINGLE_TEST defined:
//     Throws std::exception on test failure; Basic Guarantee
void test_class_RDParse2_ops(Tester & t)
{
    std::cout << "Test Suite: class RDParse2 - Operators"
              << std::endl;

    RDParse2 parser;
    bool parsed;
    bool done;
    std::string output;

    // Basic expression syntax
    parser.set("begin\nprint (3+2)\nend");
    parsed = true;
    done = true;
    output = "5\n";
    parse_it(t, parser, "Parens & operator", parsed, done, output);

    parser.set("begin\nprint 3+2)\nend");
    parsed = false;
    parse_it(t, parser, "Missing open paren", parsed);

    parser.set("begin\nprint (3+2\nend");
    parsed = false;
    parse_it(t, parser, "Missing close paren", parsed);

    parser.set("begin\nprint (3 2)\nend");
    parsed = false;
    parse_it(t, parser, "Missing op", parsed);

    parser.set("begin\nprint (3+)\nend");
    parsed = false;
    parse_it(t, parser, "No 2nd operand", parsed);

    parser.set("begin\nprint (+2)\nend");
    parsed = true;
    done = true;
    output = "2\n";
    parse_it(t, parser, "No 1st operand, not error", parsed, done, output);

    parser.set("begin\nprint (*2)\nend");
    parsed = false;
    parse_it(t, parser, "No 1st operand, error", parsed);

    // Operators +, -: values & associativity
    parser.set("begin\nprint 3+4\nend");
    parsed = true;
    done = true;
    output = "7\n";
    parse_it(t, parser, "Op+", parsed, done, output);

    parser.set("begin\nprint 3+4+5\nend");
    parsed = true;
    done = true;
    output = "12\n";
    parse_it(t, parser, "Op+ twice", parsed, done, output);

    parser.set("begin\nprint (3+4)+ 5\nend");
    parsed = true;
    done = true;
    output = "12\n";
    parse_it(t, parser, "Op+ twice, parens #1", parsed, done, output);

    parser.set("begin\nprint 3+(4+5)\nend");
    parsed = true;
    done = true;
    output = "12\n";
    parse_it(t, parser, "Op+ twice, parens #2", parsed, done, output);

    parser.set("begin\nprint 3-4\nend");
    parsed = true;
    done = true;
    output = "-1\n";
    parse_it(t, parser, "Op-", parsed, done, output);

    parser.set("begin\nprint 3-4-5\nend");
    parsed = true;
    done = true;
    output = "-6\n";
    parse_it(t, parser, "Op- twice", parsed, done, output);

    parser.set("begin\nprint (3-4)- 5\nend");
    parsed = true;
    done = true;
    output = "-6\n";
    parse_it(t, parser, "Op- twice, parens #1", parsed, done, output);

    parser.set("begin\nprint 3-(4-5)\nend");
    parsed = true;
    done = true;
    output = "4\n";
    parse_it(t, parser, "Op- twice, parens #2", parsed, done, output);

    parser.set("begin\nprint 3+4-5\nend");
    parsed = true;
    done = true;
    output = "2\n";
    parse_it(t, parser, "Op+ op-", parsed, done, output);

    parser.set("begin\nprint (3+4)- 5\nend");
    parsed = true;
    done = true;
    output = "2\n";
    parse_it(t, parser, "Op+ op-, parens #1", parsed, done, output);

    parser.set("begin\nprint 3+(4-5)\nend");
    parsed = true;
    done = true;
    output = "2\n";
    parse_it(t, parser, "Op+ op-, parens #2", parsed, done, output);

    parser.set("begin\nprint 3-4+5\nend");
    parsed = true;
    done = true;
    output = "4\n";
    parse_it(t, parser, "Op- o+", parsed, done, output);

    parser.set("begin\nprint (3-4)+ 5\nend");
    parsed = true;
    done = true;
    output = "4\n";
    parse_it(t, parser, "Op- op+, parens #1", parsed, done, output);

    parser.set("begin\nprint 3-(4+5)\nend");
    parsed = true;
    done = true;
    output = "-6\n";
    parse_it(t, parser, "Op- op+, parens #2", parsed, done, output);

    // Operators *, /: values & associativity
    parser.set("begin\nprint 3*4\nend");
    parsed = true;
    done = true;
    output = "12\n";
    parse_it(t, parser, "Op*", parsed, done, output);

    parser.set("begin\nprint 3*4*5\nend");
    parsed = true;
    done = true;
    output = "60\n";
    parse_it(t, parser, "Op* twice", parsed, done, output);

    parser.set("begin\nprint (3*4)*5\nend");
    parsed = true;
    done = true;
    output = "60\n";
    parse_it(t, parser, "Op* twice, parens #1", parsed, done, output);

    parser.set("begin\nprint 3*(4*5)\nend");
    parsed = true;
    done = true;
    output = "60\n";
    parse_it(t, parser, "Op* twice, parens #2", parsed, done, output);

    parser.set("begin\nprint 3/4\nend");
    parsed = true;
    done = true;
    output = "0.75\n";
    parse_it(t, parser, "Op/", parsed, done, output);

    parser.set("begin\nprint 3/4/5\nend");
    parsed = true;
    done = true;
    output = "0.15\n";
    parse_it(t, parser, "Op/ twice", parsed, done, output);

    parser.set("begin\nprint (3/4)/5\nend");
    parsed = true;
    done = true;
    output = "0.15\n";
    parse_it(t, parser, "Op/ twice, parens #1", parsed, done, output);

    parser.set("begin\nprint 3/(4/5)\nend");
    parsed = true;
    done = true;
    output = "3.75\n";
    parse_it(t, parser, "Op/ twice, parens #2", parsed, done, output);

    parser.set("begin\nprint 3*4/5\nend");
    parsed = true;
    done = true;
    output = "2.4\n";
    parse_it(t, parser, "Op* op/", parsed, done, output);

    parser.set("begin\nprint (3*4)/5\nend");
    parsed = true;
    done = true;
    output = "2.4\n";
    parse_it(t, parser, "Op* op/, parens #1", parsed, done, output);

    parser.set("begin\nprint 3*(4/5)\nend");
    parsed = true;
    done = true;
    output = "2.4\n";
    parse_it(t, parser, "Op* op/, parens #2", parsed, done, output);

    parser.set("begin\nprint 3/4*5\nend");
    parsed = true;
    done = true;
    output = "3.75\n";
    parse_it(t, parser, "Op/ o*", parsed, done, output);

    parser.set("begin\nprint (3/4)*5\nend");
    parsed = true;
    done = true;
    output = "3.75\n";
    parse_it(t, parser, "Op/ op*, parens #1", parsed, done, output);

    parser.set("begin\nprint 3/(4*5)\nend");
    parsed = true;
    done = true;
    output = "0.15\n";
    parse_it(t, parser, "Op/ op*, parens #2", parsed, done, output);

    // Operators ==, !=: values & associativity
    parser.set("begin\nprint 3==3\nend");
    parsed = true;
    done = true;
    output = "1\n";
    parse_it(t, parser, "Op== true", parsed, done, output);

    parser.set("begin\nprint 3==4\nend");
    parsed = true;
    done = true;
    output = "0\n";
    parse_it(t, parser, "Op== false", parsed, done, output);

    parser.set("begin\nprint 1==2==0\nend");
    parsed = true;
    done = true;
    output = "1\n";
    parse_it(t, parser, "Op== twice", parsed, done, output);

    parser.set("begin\nprint (1==2)==0\nend");
    parsed = true;
    done = true;
    output = "1\n";
    parse_it(t, parser, "Op== twice, parens #1", parsed, done, output);

    parser.set("begin\nprint 1==(2==0)\nend");
    parsed = true;
    done = true;
    output = "0\n";
    parse_it(t, parser, "Op== twice, parens #2", parsed, done, output);

    parser.set("begin\nprint 3!=4\nend");
    parsed = true;
    done = true;
    output = "1\n";
    parse_it(t, parser, "Op!= true", parsed, done, output);

    parser.set("begin\nprint 3!=3\nend");
    parsed = true;
    done = true;
    output = "0\n";
    parse_it(t, parser, "Op!= false", parsed, done, output);

    parser.set("begin\nprint 1!=2!=0\nend");
    parsed = true;
    done = true;
    output = "1\n";
    parse_it(t, parser, "Op!= twice", parsed, done, output);

    parser.set("begin\nprint (1!=2)!=0\nend");
    parsed = true;
    done = true;
    output = "1\n";
    parse_it(t, parser, "Op!= twice, parens #1", parsed, done, output);

    parser.set("begin\nprint 1!=(2!=0)\nend");
    parsed = true;
    done = true;
    output = "0\n";
    parse_it(t, parser, "Op!= twice, parens #2", parsed, done, output);

    parser.set("begin\nprint 1==2!=0\nend");
    parsed = true;
    done = true;
    output = "0\n";
    parse_it(t, parser, "Op== op!=", parsed, done, output);

    parser.set("begin\nprint (1==2)!=0\nend");
    parsed = true;
    done = true;
    output = "0\n";
    parse_it(t, parser, "Op== op!=, parens #1", parsed, done, output);

    parser.set("begin\nprint 1==(2!=0)\nend");
    parsed = true;
    done = true;
    output = "1\n";
    parse_it(t, parser, "Op== op!=, parens #2", parsed, done, output);

    parser.set("begin\nprint 1!=2==0\nend");
    parsed = true;
    done = true;
    output = "0\n";
    parse_it(t, parser, "Op!= o==", parsed, done, output);

    parser.set("begin\nprint (1!=2)==0\nend");
    parsed = true;
    done = true;
    output = "0\n";
    parse_it(t, parser, "Op!= op==, parens #1", parsed, done, output);

    parser.set("begin\nprint 1!=(2==0)\nend");
    parsed = true;
    done = true;
    output = "1\n";
    parse_it(t, parser, "Op!= op==, parens #2", parsed, done, output);

    // Precedence checks
    parser.set("begin\nprint 3+4*5\nend");
    parsed = true;
    done = true;
    output = "23\n";
    parse_it(t, parser, "Precedence op+ op*", parsed, done, output);

    parser.set("begin\nprint 3+4/5\nend");
    parsed = true;
    done = true;
    output = "3.8\n";
    parse_it(t, parser, "Precedence op+ op/", parsed, done, output);

    parser.set("begin\nprint 3+4==5\nend");
    parsed = true;
    done = true;
    output = "0\n";
    parse_it(t, parser, "Precedence op+ op==", parsed, done, output);

    parser.set("begin\nprint 3+4!=5\nend");
    parsed = true;
    done = true;
    output = "1\n";
    parse_it(t, parser, "Precedence op+ op!=", parsed, done, output);

    parser.set("begin\nprint 3-4*5\nend");
    parsed = true;
    done = true;
    output = "-17\n";
    parse_it(t, parser, "Precedence op- op*", parsed, done, output);

    parser.set("begin\nprint 3-4/5\nend");
    parsed = true;
    done = true;
    output = "2.2\n";
    parse_it(t, parser, "Precedence op- op/", parsed, done, output);

    parser.set("begin\nprint 3-4==5\nend");
    parsed = true;
    done = true;
    output = "0\n";
    parse_it(t, parser, "Precedence op- op==", parsed, done, output);

    parser.set("begin\nprint 3-4!=5\nend");
    parsed = true;
    done = true;
    output = "1\n";
    parse_it(t, parser, "Precedence op- op!=", parsed, done, output);

    parser.set("begin\nprint 3*4+5\nend");
    parsed = true;
    done = true;
    output = "17\n";
    parse_it(t, parser, "Precedence op* op+", parsed, done, output);

    parser.set("begin\nprint 3*4-5\nend");
    parsed = true;
    done = true;
    output = "7\n";
    parse_it(t, parser, "Precedence op* op-", parsed, done, output);

    parser.set("begin\nprint 3*4==5\nend");
    parsed = true;
    done = true;
    output = "0\n";
    parse_it(t, parser, "Precedence op* op==", parsed, done, output);

    parser.set("begin\nprint 3*4!=5\nend");
    parsed = true;
    done = true;
    output = "1\n";
    parse_it(t, parser, "Precedence op* op!=", parsed, done, output);

    parser.set("begin\nprint 3/4+5\nend");
    parsed = true;
    done = true;
    output = "5.75\n";
    parse_it(t, parser, "Precedence op/ op+", parsed, done, output);

    parser.set("begin\nprint 3/4-5\nend");
    parsed = true;
    done = true;
    output = "-4.25\n";
    parse_it(t, parser, "Precedence op/ op-", parsed, done, output);

    parser.set("begin\nprint 3/4==5\nend");
    parsed = true;
    done = true;
    output = "0\n";
    parse_it(t, parser, "Precedence op/ op==", parsed, done, output);

    parser.set("begin\nprint 3/4!=5\nend");
    parsed = true;
    done = true;
    output = "1\n";
    parse_it(t, parser, "Precedence op/ op!=", parsed, done, output);

    parser.set("begin\nprint 3==4+5\nend");
    parsed = true;
    done = true;
    output = "0\n";
    parse_it(t, parser, "Precedence op== op+", parsed, done, output);

    parser.set("begin\nprint 3==4-5\nend");
    parsed = true;
    done = true;
    output = "0\n";
    parse_it(t, parser, "Precedence op== op-", parsed, done, output);

    parser.set("begin\nprint 3==4*5\nend");
    parsed = true;
    done = true;
    output = "0\n";
    parse_it(t, parser, "Precedence op== op*", parsed, done, output);

    parser.set("begin\nprint 3==4/5\nend");
    parsed = true;
    done = true;
    output = "0\n";
    parse_it(t, parser, "Precedence op== op/", parsed, done, output);

    parser.set("begin\nprint 3!=4+5\nend");
    parsed = true;
    done = true;
    output = "1\n";
    parse_it(t, parser, "Precedence op!= op+", parsed, done, output);

    parser.set("begin\nprint 3!=4-5\nend");
    parsed = true;
    done = true;
    output = "1\n";
    parse_it(t, parser, "Precedence op!= op-", parsed, done, output);

    parser.set("begin\nprint 3!=4*5\nend");
    parsed = true;
    done = true;
    output = "1\n";
    parse_it(t, parser, "Precedence op!= op*", parsed, done, output);

    parser.set("begin\nprint 3!=4/5\nend");
    parsed = true;
    done = true;
    output = "1\n";
    parse_it(t, parser, "Precedence op!= op/", parsed, done, output);

    // Precedence override checks
    parser.set("begin\nprint (3+4)*5\nend");
    parsed = true;
    done = true;
    output = "35\n";
    parse_it(t, parser, "Precedence override op+ op*", parsed, done, output);

    parser.set("begin\nprint (3+4)/5\nend");
    parsed = true;
    done = true;
    output = "1.4\n";
    parse_it(t, parser, "Precedence override op+ op/", parsed, done, output);

    parser.set("begin\nprint 3+(4==5)\nend");
    parsed = true;
    done = true;
    output = "3\n";
    parse_it(t, parser, "Precedence override op+ op==", parsed, done, output);

    parser.set("begin\nprint 3+(4!=5)\nend");
    parsed = true;
    done = true;
    output = "4\n";
    parse_it(t, parser, "Precedence override op+ op!=", parsed, done, output);

    parser.set("begin\nprint (3-4)*5\nend");
    parsed = true;
    done = true;
    output = "-5\n";
    parse_it(t, parser, "Precedence override op- op*", parsed, done, output);

    parser.set("begin\nprint (3-4)/5\nend");
    parsed = true;
    done = true;
    output = "-0.2\n";
    parse_it(t, parser, "Precedence override op- op/", parsed, done, output);

    parser.set("begin\nprint 3-(4==5)\nend");
    parsed = true;
    done = true;
    output = "3\n";
    parse_it(t, parser, "Precedence override op- op==", parsed, done, output);

    parser.set("begin\nprint 3-(4!=5)\nend");
    parsed = true;
    done = true;
    output = "2\n";
    parse_it(t, parser, "Precedence override op- op!=", parsed, done, output);

    parser.set("begin\nprint 3*(4+5)\nend");
    parsed = true;
    done = true;
    output = "27\n";
    parse_it(t, parser, "Precedence override op* op+", parsed, done, output);

    parser.set("begin\nprint 3*(4-5)\nend");
    parsed = true;
    done = true;
    output = "-3\n";
    parse_it(t, parser, "Precedence override op* op-", parsed, done, output);

    parser.set("begin\nprint 3*(4==5)\nend");
    parsed = true;
    done = true;
    output = "0\n";
    parse_it(t, parser, "Precedence override op* op==", parsed, done, output);

    parser.set("begin\nprint 3*(4!=5)\nend");
    parsed = true;
    done = true;
    output = "3\n";
    parse_it(t, parser, "Precedence override op* op!=", parsed, done, output);

    parser.set("begin\nprint 3/(4+6)\nend");
    parsed = true;
    done = true;
    output = "0.3\n";
    parse_it(t, parser, "Precedence override op/ op+", parsed, done, output);

    parser.set("begin\nprint 3/(4-5)\nend");
    parsed = true;
    done = true;
    output = "-3\n";
    parse_it(t, parser, "Precedence override op/ op-", parsed, done, output);

    parser.set("begin\nprint 3/(4==4)\nend");
    parsed = true;
    done = true;
    output = "3\n";
    parse_it(t, parser, "Precedence override op/ op==", parsed, done, output);

    parser.set("begin\nprint 3/(4!=5)\nend");
    parsed = true;
    done = true;
    output = "3\n";
    parse_it(t, parser, "Precedence override op/ op!=", parsed, done, output);

    parser.set("begin\nprint (3==4)+ 5\nend");
    parsed = true;
    done = true;
    output = "5\n";
    parse_it(t, parser, "Precedence override op== op+", parsed, done, output);

    parser.set("begin\nprint (3==4)- 5\nend");
    parsed = true;
    done = true;
    output = "-5\n";
    parse_it(t, parser, "Precedence override op== op-", parsed, done, output);

    parser.set("begin\nprint (3==4)*5\nend");
    parsed = true;
    done = true;
    output = "0\n";
    parse_it(t, parser, "Precedence override op== op*", parsed, done, output);

    parser.set("begin\nprint (3==4)/5\nend");
    parsed = true;
    done = true;
    output = "0\n";
    parse_it(t, parser, "Precedence override op== op/", parsed, done, output);

    parser.set("begin\nprint (3!=4)+ 5\nend");
    parsed = true;
    done = true;
    output = "6\n";
    parse_it(t, parser, "Precedence override op!= op+", parsed, done, output);

    parser.set("begin\nprint (3!=4)- 5\nend");
    parsed = true;
    done = true;
    output = "-4\n";
    parse_it(t, parser, "Precedence override op!= op-", parsed, done, output);

    parser.set("begin\nprint (3!=4)*5\nend");
    parsed = true;
    done = true;
    output = "5\n";
    parse_it(t, parser, "Precedence override op!= op*", parsed, done, output);

    parser.set("begin\nprint (3!=4)/5\nend");
    parsed = true;
    done = true;
    output = "0.2\n";
    parse_it(t, parser, "Precedence override op!= op/", parsed, done, output);
}


// test_class_RDParse2_vars
// Test suite for class RDParse2, variables
// Pre: None.
// Post:
//     Pass/fail status of tests have been registered with t.
//     Appropriate messages have been printed to cout.
// If SINGLE_TEST not defined:
//     Does not throw (No-Throw Guarantee)
// If SINGLE_TEST defined:
//     Throws std::exception on test failure; Basic Guarantee
void test_class_RDParse2_vars(Tester & t)
{
    std::cout << "Test Suite: class RDParse2 - variables"
              << std::endl;

    RDParse2 parser;
    bool parsed;
    bool done;
    std::string output;

    parser.set("begin\na=7\nprint a\nend");
    parsed = true;
    done = true;
    output = "7\n";
    parse_it(t, parser, "Used var", parsed, done, output);

    parser.set("begin\na=7\nprint 3\nend");
    parsed = true;
    done = true;
    output = "3\n";
    parse_it(t, parser, "Unused var", parsed, done, output);

    parser.set("begin\na=7\nb=2\nprint a\nend");
    parsed = true;
    done = true;
    output = "7\n";
    parse_it(t, parser, "2 vars, print 1st", parsed, done, output);

    parser.set("begin\na=7\nb=2\nprint b\nend");
    parsed = true;
    done = true;
    output = "2\n";
    parse_it(t, parser, "2 vars, print 2nd", parsed, done, output);

    parser.set("begin\na=7\nb=a\nprint b\nend");
    parsed = true;
    done = true;
    output = "7\n";
    parse_it(t, parser, "var=var", parsed, done, output);

    parser.set("begin\na=7\na=6\nprint a\nend");
    parsed = true;
    done = true;
    output = "6\n";
    parse_it(t, parser, "Reassign var", parsed, done, output);

    parser.set("begin\na=7\nprint a\na=6\nprint a\nend");
    parsed = true;
    done = true;
    output = "7\n6\n";
    parse_it(t, parser, "Print var, change, print", parsed, done, output);

    parser.set("begin\na=7\na=a\nprint a\nend");
    parsed = true;
    done = true;
    output = "7\n";
    parse_it(t, parser, "Var=self", parsed, done, output);

    parser.set("begin\na=7\nb=a+1\nprint b\nend");
    parsed = true;
    done = true;
    output = "8\n";
    parse_it(t, parser, "Var=func of var", parsed, done, output);

    parser.set("begin\na=7\na=a+1\nprint a\nend");
    parsed = true;
    done = true;
    output = "8\n";
    parse_it(t, parser, "Var=func of self", parsed, done, output);

    parser.set("begin\na=7\na=(a+1)+(a+2)\nprint a\nend");
    parsed = true;
    done = true;
    output = "17\n";
    parse_it(t, parser, "Var=func+func of self", parsed, done, output);

    parser.set("begin\na=7\nb=2\nc=a+b\nprint c\nend");
    parsed = true;
    done = true;
    output = "9\n";
    parse_it(t, parser, "Var=var+var", parsed, done, output);

    parser.set("begin\na=7\nb=2\nb=a+b\nprint b\nend");
    parsed = true;
    done = true;
    output = "9\n";
    parse_it(t, parser, "Var=var+self", parsed, done, output);
}


// test_class_RDParse2_long
// Test suite for class RDParse2, long programs
// Pre: None.
// Post:
//     Pass/fail status of tests have been registered with t.
//     Appropriate messages have been printed to cout.
// If SINGLE_TEST not defined:
//     Does not throw (No-Throw Guarantee)
// If SINGLE_TEST defined:
//     Throws std::exception on test failure; Basic Guarantee
void test_class_RDParse2_long(Tester & t)
{
    std::cout << "Test Suite: class RDParse2 - long programs"
              << std::endl;

    RDParse2 parser;
    bool parsed;
    bool done;
    std::string output;
    std::string prog;

    const int BASENUMPARENS = 50;  // Was 1000, but some got stack
                                   //  overflow
    std::string open(BASENUMPARENS,'(');
    std::string close(BASENUMPARENS,')');

    parser.set("begin\nprint "+open+"42"+close+"\n end");
    parsed = true;
    done = true;
    output = "42\n";
    parse_it(t, parser, "Many parens: matched", parsed, done, output);

    parser.set("begin\nprint "+open+"(42"+close+"\n end");
    parsed = false;
    parse_it(t, parser, "Many parens: unmatched #1", parsed);

    parser.set("begin\nprint "+open+"((42"+close+"\n end");
    parsed = false;
    parse_it(t, parser, "Many parens: unmatched #2", parsed);

    parser.set("begin\nprint "+open+"42)"+close+"\n end");
    parsed = false;
    parse_it(t, parser, "Many parens: unmatched #3", parsed);

    const int NUMPRINTS = 100;
    prog = "begin\n";
    for (int i = 0; i < NUMPRINTS; ++i)
        prog += "print 42\n";
    prog += "end";
    parser.set(prog);
    parsed = done;
    output = "";
    for (int i = 0; i < NUMPRINTS; ++i)
        output += "42\n";
    parse_it(t, parser, "Many prints: same", parsed, done, output);

    std::ostringstream oss1, oss2;
    oss1 << "begin\n";
    for (int i = 0; i < NUMPRINTS; ++i)
        oss1 << "print " << i << "\n";
    oss1 << "end";
    parser.set(oss1.str());
    parsed = done;
    for (int i = 0; i < NUMPRINTS; ++i)
         oss2 << i << "\n";
    output = oss2.str();
    parse_it(t, parser, "Many prints: different", parsed, done, output);

    const int NUMINCREMENTS = 10000;
    std::ostringstream oss3;
    prog = "begin\na = 0\n";
    for (int i = 0; i < NUMINCREMENTS; ++i)
        prog += "a = a+1\n";
    prog += "print a\nend";
    parser.set(prog);
    parsed = true;
    done = true;
    oss3 << NUMINCREMENTS << std::endl;
    output = oss3.str();
    parse_it(t, parser, "Many increments", parsed, done, output);
}


// test_class_RDParse2_complex
// Test suite for class RDParse2, more complex computations
// Pre: None.
// Post:
//     Pass/fail status of tests have been registered with t.
//     Appropriate messages have been printed to cout.
// If SINGLE_TEST not defined:
//     Does not throw (No-Throw Guarantee)
// If SINGLE_TEST defined:
//     Throws std::exception on test failure; Basic Guarantee
void test_class_RDParse2_complex(Tester & t)
{
    std::cout << "Test Suite: class RDParse2 - more complex computations"
              << std::endl;

    RDParse2 parser;
    bool parsed;
    bool done;
    double val;

    parser.set("begin\npi=0\nd=1\npi=pi+3*d\nd=d/10\npi=pi+1*d\nd=d/10\n"
               "pi=pi+4*d\n""d=d/10\npi=pi+1*d\nd=d/10\npi=pi+5*d\n"
               "d=d/10\npi=pi+9*d\nprint pi\nend");
    parsed = true;
    done = true;
    val = 3.14159;
    parse_it_precision(t, parser, "Add decimals", parsed, done, val);

    parser.set("begin\ni=0\nd=1\ne=d\n"
               "i=i+1\nd=d/i\ne=e+d\n"
               "i=i+1\nd=d/i\ne=e+d\n"
               "i=i+1\nd=d/i\ne=e+d\n"
               "i=i+1\nd=d/i\ne=e+d\n"
               "i=i+1\nd=d/i\ne=e+d\n"
               "i=i+1\nd=d/i\ne=e+d\n"
               "i=i+1\nd=d/i\ne=e+d\n"
               "i=i+1\nd=d/i\ne=e+d\n"
               "i=i+1\nd=d/i\ne=e+d\n"
               "print e\nend");
    parsed = true;
    done = true;
    val = 2.71828182845;
    parse_it_precision(t, parser, "Approx e #1", parsed, done, val);

    parser.set("begin\ne=((((((((1/9+1)/8+1)/7+1)"
               "/6+1)/5+1)/4+1)/3+1)/2+1)/1+1\n"
               "print e\nend");
    parsed = true;
    done = true;
    val = 2.71828182845;
    parse_it_precision(t, parser, "Approx e #2", parsed, done, val);

    parser.set("begin\na=((3!=54*-9.8)+ 2+1)/(4*3+(2==5-3))\n"
               "_b34____23kw=a+(4-3==1)/13+(4-2)*a\n"
               "print _b34____23kw\nend");
    parsed = true;
    done = true;
    val = 1;
    parse_it_precision(t, parser, "Messy computation", parsed, done, val);
}


// test_class_RDParse2
// Test suite for class RDParse2
// Uses other test-suite functions
// Pre: None.
// Post:
//     Pass/fail status of tests have been registered with t.
//     Appropriate messages have been printed to cout.
// If SINGLE_TEST not defined:
//     Does not throw (No-Throw Guarantee)
// If SINGLE_TEST defined:
//     Throws std::exception on test failure; Basic Guarantee
void test_class_RDParse2(Tester & t)
{
    // Do all the test suites
    std::cout << "TEST SUITES FOR CLASS RDParse2" << std::endl;
    test_class_RDParse2_simple(t);
    test_class_RDParse2_ops(t);
    test_class_RDParse2_vars(t);
    test_class_RDParse2_long(t);
    test_class_RDParse2_complex(t);
}


// ************************************************************************
// Main program
// ************************************************************************


// main
// Runs class Lex2 test suite, prints results to cout.
int main()
{
    Tester t;
#ifdef SINGLE_TEST
    try
    {
        test_class_RDParse2(t);
    }
    catch (std::exception & e)
    { /* Nothing to do here ... */ }
#else //#ifdef SINGLE_TEST
    test_class_RDParse2(t);
#endif //#ifdef SINGLE_TEST

    std::cout << std::endl;
    if (t.allPassed())
    {
        std::cout << "All tests successful" 
                  << std::endl;
    }
    else
    {
        std::cout << "Tests ********** UNSUCCESSFUL **********"
                  << std::endl;
#ifdef SINGLE_TEST
        std::cout << std::endl;
        std::cout << "NOTE: THIS PROGRAM TERMINATES AFTER THE FIRST"
                  << std::endl;
        std::cout << "UNSUCCESSFUL TEST (IF ANY). TO CHANGE THIS"
                  << std::endl;
        std::cout << "BEHAVIOR, COMMENT OUT THE LINE DEFINING THE"
                  << std::endl;
        std::cout << "CONSTANT \"SINGLE_TEST\"."
                  << std::endl;
#endif //#ifdef SINGLE_TEST
    }
    std::cout << std::endl;

    std::cout << "Press ENTER to quit ";
    while (std::cin.get() != '\n') ;

    return 0;
}

